@@ -1,4 +1,5 @@
 #include "catch.hpp"
+#include <iostream>
 
 #include <initializer_list>
 
@@ -49,10 +50,14 @@ TEST_CASE("Basic options", "[options]")
   options.add_options()
     ("long", "a long option")
     ("s,short", "a short option")
+    ("quick,brown", "An option with multiple long names and no short name")
+    ("f,ox,jumped", "An option with multiple long names and a short name")
+    ("over,z,lazy,dog", "An option with multiple long names and a short name, not listed first")
     ("value", "an option with a value", cxxopts::value<std::string>())
     ("a,av", "a short option with a value", cxxopts::value<std::string>())
     ("6,six", "a short number option")
     ("p, space", "an option with space between short and long")
+    ("period.delimited", "an option with a period in the long name")
     ("nothing", "won't exist", cxxopts::value<std::string>())
     ;
 
@@ -67,6 +72,15 @@ TEST_CASE("Basic options", "[options]")
     "-6",
     "-p",
     "--space",
+    "--quick",
+    "--ox",
+    "-f",
+    "--brown",
+    "-z",
+    "--over",
+    "--dog",
+    "--lazy",
+    "--period.delimited",
   });
 
   auto** actual_argv = argv.argv();
@@ -83,9 +97,13 @@ TEST_CASE("Basic options", "[options]")
   CHECK(result.count("6") == 1);
   CHECK(result.count("p") == 2);
   CHECK(result.count("space") == 2);
+  CHECK(result.count("quick") == 2);
+  CHECK(result.count("f") == 2);
+  CHECK(result.count("z") == 4);
+  CHECK(result.count("period.delimited") == 1);
 
   auto& arguments = result.arguments();
-  REQUIRE(arguments.size() == 7);
+  REQUIRE(arguments.size() == 16);
   CHECK(arguments[0].key() == "long");
   CHECK(arguments[0].value() == "true");
   CHECK(arguments[0].as<bool>() == true);
@@ -94,7 +112,9 @@ TEST_CASE("Basic options", "[options]")
   CHECK(arguments[2].key() == "value");
   CHECK(arguments[3].key() == "av");
 
-  CHECK_THROWS_AS(result["nothing"].as<std::string>(), cxxopts::option_has_no_value_exception&);
+  CHECK_THROWS_AS(result["nothing"].as<std::string>(), cxxopts::exceptions::option_has_no_value&);
+
+  CHECK(options.program() == "tester");
 }
 
 TEST_CASE("Short options", "[options]")
@@ -102,9 +122,11 @@ TEST_CASE("Short options", "[options]")
   cxxopts::Options options("test_short", " - test short options");
 
   options.add_options()
-    ("a", "a short option", cxxopts::value<std::string>());
+    ("a", "a short option", cxxopts::value<std::string>())
+    ("b", "b option")
+    ("c", "c option", cxxopts::value<std::string>());
 
-  Argv argv({"test_short", "-a", "value"});
+  Argv argv({"test_short", "-a", "value", "-bcfoo=something"});
 
   auto actual_argv = argv.argv();
   auto argc = argv.argc();
@@ -114,8 +136,18 @@ TEST_CASE("Short options", "[options]")
   CHECK(result.count("a") == 1);
   CHECK(result["a"].as<std::string>() == "value");
 
+  auto& arguments = result.arguments();
+  REQUIRE(arguments.size() == 3);
+  CHECK(arguments[0].key() == "a");
+  CHECK(arguments[0].value() == "value");
+
+  CHECK(result.count("b") == 1);
+  CHECK(result.count("c") == 1);
+
+  CHECK(result["c"].as<std::string>() == "foo=something");
+
   REQUIRE_THROWS_AS(options.add_options()("", "nothing option"),
-    cxxopts::invalid_option_format_error&);
+    cxxopts::exceptions::invalid_option_format&);
 }
 
 TEST_CASE("No positional", "[positional]")
@@ -228,7 +260,7 @@ TEST_CASE("Positional not valid", "[positional]") {
   auto** argv = av.argv();
   auto argc = av.argc();
 
-  CHECK_THROWS_AS(options.parse(argc, argv), cxxopts::option_not_exists_exception&);
+  CHECK_THROWS_AS(options.parse(argc, argv), cxxopts::exceptions::no_such_option&);
 }
 
 TEST_CASE("Positional with empty arguments", "[positional]") {
@@ -287,7 +319,7 @@ TEST_CASE("Boolean without implicit value", "[implicit]")
     auto** argv = av.argv();
     auto argc = av.argc();
 
-    CHECK_THROWS_AS(options.parse(argc, argv), cxxopts::missing_argument_exception&);
+    CHECK_THROWS_AS(options.parse(argc, argv), cxxopts::exceptions::missing_argument&);
   }
 
   SECTION("With equal-separated true") {
@@ -375,10 +407,12 @@ TEST_CASE("Default values", "[default]")
 TEST_CASE("Parse into a reference", "[reference]")
 {
   int value = 0;
+  bool b_value = true;
 
   cxxopts::Options options("into_reference", "parses into a reference");
   options.add_options()
-    ("ref", "A reference", cxxopts::value(value));
+    ("ref", "A reference", cxxopts::value(value))
+    ("bool", "A bool", cxxopts::value(b_value));
 
   Argv av({"into_reference", "--ref", "42"});
 
@@ -388,6 +422,8 @@ TEST_CASE("Parse into a reference", "[reference]")
   auto result = options.parse(argc, argv);
   CHECK(result.count("ref") == 1);
   CHECK(value == 42);
+  CHECK(result.count("bool") == 0);
+  CHECK(b_value == true);
 }
 
 TEST_CASE("Integers", "[options]")
@@ -453,7 +489,7 @@ TEST_CASE("Unsigned integers", "[options]")
   auto argc = av.argc();
 
   options.parse_positional("positional");
-  CHECK_THROWS_AS(options.parse(argc, argv), cxxopts::argument_incorrect_type&);
+  CHECK_THROWS_AS(options.parse(argc, argv), cxxopts::exceptions::incorrect_argument_type&);
 }
 
 TEST_CASE("Integer bounds", "[integer]")
@@ -490,12 +526,12 @@ TEST_CASE("Overflow on boundary", "[integer]")
   int8_t si;
   uint8_t ui;
 
-  CHECK_THROWS_AS((integer_parser("128", si)), cxxopts::argument_incorrect_type&);
-  CHECK_THROWS_AS((integer_parser("-129", si)), cxxopts::argument_incorrect_type&);
-  CHECK_THROWS_AS((integer_parser("256", ui)), cxxopts::argument_incorrect_type&);
-  CHECK_THROWS_AS((integer_parser("-0x81", si)), cxxopts::argument_incorrect_type&);
-  CHECK_THROWS_AS((integer_parser("0x80", si)), cxxopts::argument_incorrect_type&);
-  CHECK_THROWS_AS((integer_parser("0x100", ui)), cxxopts::argument_incorrect_type&);
+  CHECK_THROWS_AS((integer_parser("128", si)), cxxopts::exceptions::incorrect_argument_type&);
+  CHECK_THROWS_AS((integer_parser("-129", si)), cxxopts::exceptions::incorrect_argument_type&);
+  CHECK_THROWS_AS((integer_parser("256", ui)), cxxopts::exceptions::incorrect_argument_type&);
+  CHECK_THROWS_AS((integer_parser("-0x81", si)), cxxopts::exceptions::incorrect_argument_type&);
+  CHECK_THROWS_AS((integer_parser("0x80", si)), cxxopts::exceptions::incorrect_argument_type&);
+  CHECK_THROWS_AS((integer_parser("0x100", ui)), cxxopts::exceptions::incorrect_argument_type&);
 }
 
 TEST_CASE("Integer overflow", "[options]")
@@ -512,11 +548,11 @@ TEST_CASE("Integer overflow", "[options]")
   auto argc = av.argc();
 
   options.parse_positional("positional");
-  CHECK_THROWS_AS(options.parse(argc, argv), cxxopts::argument_incorrect_type&);
+  CHECK_THROWS_AS(options.parse(argc, argv), cxxopts::exceptions::incorrect_argument_type&);
 
   int integer = 0;
-  CHECK_THROWS_AS((integer_parser("23423423423", integer)), cxxopts::argument_incorrect_type&);
-  CHECK_THROWS_AS((integer_parser("234234234234", integer)), cxxopts::argument_incorrect_type&);
+  CHECK_THROWS_AS((integer_parser("23423423423", integer)), cxxopts::exceptions::incorrect_argument_type&);
+  CHECK_THROWS_AS((integer_parser("234234234234", integer)), cxxopts::exceptions::incorrect_argument_type&);
 }
 
 TEST_CASE("Floats", "[options]")
@@ -557,7 +593,7 @@ TEST_CASE("Invalid integers", "[integer]") {
     auto argc = av.argc();
 
     options.parse_positional("positional");
-    CHECK_THROWS_AS(options.parse(argc, argv), cxxopts::argument_incorrect_type&);
+    CHECK_THROWS_AS(options.parse(argc, argv), cxxopts::exceptions::incorrect_argument_type&);
 }
 
 TEST_CASE("Booleans", "[boolean]") {
@@ -663,7 +699,7 @@ TEST_CASE("Unrecognised options", "[options]") {
   auto argc = av.argc();
 
   SECTION("Default behaviour") {
-    CHECK_THROWS_AS(options.parse(argc, argv), cxxopts::option_not_exists_exception&);
+    CHECK_THROWS_AS(options.parse(argc, argv), cxxopts::exceptions::no_such_option&);
   }
 
   SECTION("After allowing unrecognised options") {
@@ -682,22 +718,22 @@ TEST_CASE("Allow bad short syntax", "[options]") {
     ("s,short", "a short option");
 
   Argv av({
-    "unknown_options",
-    "-some_bad_short",
+    "--ab?",
+    "-?b?#@"
   });
 
   auto** argv = av.argv();
   auto argc = av.argc();
 
   SECTION("Default behaviour") {
-    CHECK_THROWS_AS(options.parse(argc, argv), cxxopts::option_syntax_exception&);
+    CHECK_THROWS_AS(options.parse(argc, argv), cxxopts::exceptions::invalid_option_syntax&);
   }
 
   SECTION("After allowing unrecognised options") {
     options.allow_unrecognised_options();
     CHECK_NOTHROW(options.parse(argc, argv));
     REQUIRE(argc == 2);
-    CHECK_THAT(argv[1], Catch::Equals("-some_bad_short"));
+    CHECK_THAT(argv[1], Catch::Equals("-?b?#@"));
   }
 }
 
@@ -713,7 +749,7 @@ TEST_CASE("Invalid option syntax", "[options]") {
   auto argc = av.argc();
 
   SECTION("Default behaviour") {
-    CHECK_THROWS_AS(options.parse(argc, argv), cxxopts::option_syntax_exception&);
+    CHECK_THROWS_AS(options.parse(argc, argv), cxxopts::exceptions::invalid_option_syntax&);
   }
 }
 
@@ -732,7 +768,7 @@ TEST_CASE("Options empty", "[options]") {
   auto** argv = argv_.argv();
 
   CHECK(options.groups().empty());
-  CHECK_THROWS_AS(options.parse(argc, argv), cxxopts::option_not_exists_exception&);
+  CHECK_THROWS_AS(options.parse(argc, argv), cxxopts::exceptions::no_such_option&);
 }
 
 TEST_CASE("Initializer list with group", "[options]") {
@@ -779,24 +815,30 @@ TEST_CASE("Option add with add_option(string, Option)", "[options]") {
 
   options.add_option("", option_1);
   options.add_option("TEST", {"a,aggregate", "test option 2", cxxopts::value<int>(), "AGGREGATE"});
+  options.add_option("TEST", {"multilong,m,multilong-alias", "test option 3", cxxopts::value<int>(), "An option with multiple long names"});
 
   Argv argv_({
        "test",
        "--test",
        "5",
        "-a",
-       "4"
+       "4",
+       "--multilong-alias",
+       "6"
      });
   auto argc = argv_.argc();
   auto** argv = argv_.argv();
   auto result = options.parse(argc, argv);
 
-  CHECK(result.arguments().size()==2);
+  CHECK(result.arguments().size() == 3);
   CHECK(options.groups().size() == 2);
   CHECK(result.count("address") == 0);
   CHECK(result.count("aggregate") == 1);
   CHECK(result.count("test") == 1);
   CHECK(result["aggregate"].as<int>() == 4);
+  CHECK(result["multilong"].as<int>() == 6);
+  CHECK(result["multilong-alias"].as<int>() == 6);
+  CHECK(result["m"].as<int>() == 6);
   CHECK(result["test"].as<int>() == 5);
 }
 
@@ -830,3 +872,107 @@ TEST_CASE("Parameter follow option", "[parameter]") {
   CHECK(job_values[2] == 10);
   CHECK(job_values[3] == 5);
 }
+
+TEST_CASE("Iterator", "[iterator]") {
+  cxxopts::Options options("tester", " - test iterating over parse result");
+
+  options.add_options()
+    ("long", "a long option")
+    ("s,short", "a short option")
+    ("a", "a short-only option")
+    ("value", "an option with a value", cxxopts::value<std::string>())
+    ("default", "an option with default value", cxxopts::value<int>()->default_value("42"))
+    ("nothing", "won't exist", cxxopts::value<std::string>())
+    ;
+
+  Argv argv({
+    "tester",
+    "--long",
+    "-s",
+    "-a",
+    "--value",
+    "value",
+  });
+
+  auto** actual_argv = argv.argv();
+  auto argc = argv.argc();
+
+  auto result = options.parse(argc, actual_argv);
+
+  auto iter = result.begin();
+
+  REQUIRE(iter != result.end());
+  CHECK(iter->key() == "long");
+  CHECK(iter->value() == "true");
+
+  REQUIRE(++iter != result.end());
+  CHECK(iter->key() == "short");
+  CHECK(iter->value() == "true");
+
+  REQUIRE(++iter != result.end());
+  CHECK(iter->key() == "a");
+  CHECK(iter->value() == "true");
+
+  REQUIRE(++iter != result.end());
+  CHECK(iter->key() == "value");
+  CHECK(iter->value() == "value");
+
+  REQUIRE(++iter != result.end());
+  CHECK(iter->key() == "default");
+  CHECK(iter->value() == "42");
+  
+  REQUIRE(++iter == result.end());
+}
+
+TEST_CASE("Iterator no args", "[iterator]") {
+  cxxopts::Options options("tester", " - test iterating over parse result");
+
+  options.add_options()
+    ("value", "an option with a value", cxxopts::value<std::string>())
+    ("default", "an option with default value", cxxopts::value<int>()->default_value("42"))
+    ("nothing", "won't exist", cxxopts::value<std::string>())
+    ;
+
+  Argv argv({
+    "tester",
+  });
+
+  auto** actual_argv = argv.argv();
+  auto argc = argv.argc();
+
+  auto result = options.parse(argc, actual_argv);
+
+  auto iter = result.begin();
+
+  REQUIRE(iter != result.end());
+  CHECK(iter->key() == "default");
+  CHECK(iter->value() == "42");
+  
+  ++iter;
+  CHECK(iter == result.end());
+}
+
+
+TEST_CASE("No Options help", "[options]")
+{
+  std::vector<std::string> positional;
+
+  cxxopts::Options options("test", "test no options help");
+
+  // explicitly setting custom help empty to overwrite
+  // default "[OPTION...]" when there are no options
+  options.positional_help("<posArg1>...<posArgN>")
+    .custom_help("")
+    .add_options()
+      ("positional", "", cxxopts::value<std::vector<std::string>>(positional));
+
+  Argv av({"test", "posArg1", "posArg2", "posArg3"});
+
+  auto argc   = av.argc();
+  auto** argv = av.argv();
+
+  options.parse_positional({"positional"});
+
+  CHECK_NOTHROW(options.parse(argc, argv));
+  CHECK(options.help().find("test <posArg1>...<posArgN>") != std::string::npos);
+}
